@@ -7,6 +7,7 @@
 #include <fcntl.h>
 #include <inttypes.h>
 #include <libgen.h>
+#include <poll.h>
 #include <pthread.h>
 #include <sched.h>
 #include <stdarg.h>
@@ -242,12 +243,6 @@ __lxcfs_fuse_ops int proc_access(const char *path, int mask)
 	return 0;
 }
 
-__lxcfs_fuse_ops int proc_release(const char *path, struct fuse_file_info *fi)
-{
-	do_release_file_info(fi);
-	return 0;
-}
-
 /**
  * Gets a non-hierarchical memory controller limit, or UINT64_MAX if no limit is
  * in place. If `swap` is true, reads 'swap' (v2) or 'memsw' (v1); otherwise
@@ -999,6 +994,8 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 	char *cache = d->buf + CPUALL_MAX_SIZE;
 	size_t cache_size = d->buflen - CPUALL_MAX_SIZE;
 	int cg_cpu_usage_size = 0;
+	bool use_view;
+	int max_cpus = 0;
 
 	if (offset) {
 		size_t left;
@@ -1064,6 +1061,13 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 		lxcfs_v("proc_stat_read failed to read from cpuacct, falling back to the host's /proc/stat");
 	}
 
+	if (cgroup_ops->can_use_cpuview(cgroup_ops) && opts && opts->use_cfs)
+		use_view = true;
+	else
+		use_view = false;
+	if (use_view)
+		max_cpus = max_cpu_count(cg, cpu_cg);
+
 	while (getline(&line, &linelen, f) != -1) {
 		ssize_t l;
 		char cpu_char[10]; /* That's a lot of cores */
@@ -1091,12 +1095,15 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 		if (sscanf(cpu_char, "%d", &physcpu) != 1)
 			continue;
 
+		if (use_view && max_cpus > 0 && (curcpu + 1) == max_cpus)
+			continue;	// cannot break here because we need to consume all non-cpu lines
+
 		if (!cpu_in_cpuset(physcpu, cpuset))
 			continue;
 
 		curcpu++;
 
-		if (cgroup_ops->can_use_cpuview(cgroup_ops) && opts && opts->use_cfs)
+		if (use_view)
 			cpu_to_render = curcpu;
 		else
 			cpu_to_render = physcpu;
@@ -1710,3 +1717,23 @@ __lxcfs_fuse_ops int proc_read(const char *path, char *buf, size_t size,
 
 	return -EINVAL;
 }
+
+__lxcfs_fuse_ops int proc_write(const char *path, const char *buf, size_t size,
+				off_t offset, struct fuse_file_info *fi)
+{
+	return -EINVAL;
+}
+
+__lxcfs_fuse_ops int proc_poll(const char *path, struct fuse_file_info *fi,
+			       struct fuse_pollhandle *ph, unsigned *reventsp)
+{
+	fuse_pollhandle_destroy(ph);
+	*reventsp = DEFAULT_POLLMASK;
+	return 0;
+}
+
+__lxcfs_fuse_ops int proc_release(const char *path, struct fuse_file_info *fi)
+{
+	do_release_file_info(fi);
+	return 0;
+}
